/* pmx/model.fsh */
#ifdef GL_ES
precision highp float;
#endif
#if __VERSION__ < 130
#define in varying
#define outPixelColor gl_FragColor
#define texture(samp, uv) texture2D((samp), (uv))
#else
out vec4 outPixelColor;
#endif
uniform bool hasDepthTexture;

uniform sampler2D mainTexture;
uniform sampler2D toonTexture;
uniform sampler2D sphereTexture;
uniform sampler2D depthTexture;

uniform vec4 mainTextureBlend;
uniform vec4 sphereTextureBlend;
uniform vec4 toonTextureBlend;
uniform vec2 depthTextureSize;

uniform vec3 lightDirection;
uniform vec3 lightColor;

uniform vec3 diffuseColor;
//uniform vec3 ambientColor;
uniform vec3 specularColor;
uniform float materialShininess;

uniform float opacity;

in vec2 outMainTexCoord;
in vec2 outSphereTexCoord;
in vec2 outToonTexCoord;
in vec4 outShadowTexCoord;
in vec4 outDepth;
//in vec4 outUVA1;
in vec3 outNormal;
in vec3 outEyeNormal;
in vec3 outEyeView;

const float kOne = 1.0;
const float kZero = 0.0;
const vec4 kOne4 = vec4(kOne, kOne, kOne, kOne);
const vec3 kZero3 = vec3(kZero, kZero, kZero);
const vec3 kOne3 = vec3(kOne, kOne, kOne);

vec2 calcuSphereMap(const vec3 normal) {
    return 0.5 * (1.0 + vec2(1.0, -1.0) * normal.xy);
}

vec2 calcuToon(const vec3 normal) {
    return vec2(0.0, 0.5 * (1.0 - dot(lightDirection, normal)));
}

vec3 calcuSpecular(const vec3 cameraDirection, const vec3 normal) {
    vec3 harfAngle = normalize(lightDirection + cameraDirection);
    float specularWeight = pow(max(0.001, dot(harfAngle, normal)), materialShininess);
    return specularWeight * specularColor;
}

vec4 calcuTexture(vec3 normal) {
    vec4 textureColor = kOne4;
    
#ifdef HAS_MAIN_TEXTURE
        float alpha = mainTextureBlend.a;
        textureColor = texture(mainTexture, outMainTexCoord);
        textureColor.rgb *= mainTextureBlend.rgb * alpha + (kOne - alpha);
#endif
    
#ifdef HAS_SPHERE_TEXTURE
        vec2 texCoord = calcuSphereMap(normal);
#ifdef IS_SPH_TEXTURE
            textureColor.rgb *= texture(sphereTexture, texCoord).rgb;
#endif
#ifdef IS_SPA_TEXTURE
            textureColor.rgb += texture(sphereTexture, texCoord).rgb;
#endif
#ifdef IS_SUB_TEXTURE
            textureColor.rgb *= texture(sphereTexture, outUVA1.xy).rgb;
#endif
        textureColor.rgb *= sphereTextureBlend.rgb;
#endif
    
    return textureColor;
}

vec4 calcuColor(const vec3 cameraDirection, const vec3 normal) {
    vec4 color;
    
#ifdef HAS_SPECULAR
    vec3 specular = calcuSpecular(cameraDirection, normal);
#if defined(HAS_MAIN_TEXTURE) || defined(HAS_SPHERE_TEXTURE)
    vec4 texColor = calcuTexture(normal);
    color.rgb = lightColor * min(diffuseColor * texColor.rgb + specular, texColor.rgb);
#else
    color.rgb = lightColor * diffuseColor + specular;
#endif
#else
#if defined(HAS_MAIN_TEXTURE) || defined(HAS_SPHERE_TEXTURE)
    vec4 texColor = calcuTexture(normal);
    color.rgb = lightColor * min(diffuseColor * texColor.rgb, texColor.rgb);
#else
    color.rgb = lightColor * diffuseColor;
#endif
#endif
    
#if defined(HAS_MAIN_TEXTURE) || defined(HAS_SPHERE_TEXTURE)
    color.a = texColor.a * opacity;
#else
    color.a = opacity;
#endif
    
    return color;
}

const vec4 unpackMask = vec4(1.0, 1.0 / 255.0, 1.0 / (255.0 * 255.0), 1.0 / (255.0 * 255.0 * 255.0));

float unpack_depth(const vec4 RGBA)
{
    return dot(RGBA, unpackMask);
}

void main() {
    vec3 eyeNormal = normalize(outEyeNormal);
    vec3 eyeView = normalize(outEyeView);
    
    vec4 color = calcuColor(eyeView, eyeNormal);

    // Toon Texture
#ifdef HAS_TOON_TEXTURE
        vec2 toonCoord;
#ifdef HAS_DEPTH_TEXTURE
            if (0.0 < outDepth.w) {
                vec4 rgbaDepth = texture(depthTexture, outShadowTexCoord.xy / outShadowTexCoord.w);
                float depth = unpack_depth(rgbaDepth);
                float linerDepth = length(outDepth);
                if(linerDepth > depth){
                    toonCoord = vec2(kZero, kOne);
                } else {
                    toonCoord = vec2(kZero, kZero);
                }
            } else {
                toonCoord = vec2(kZero, kZero);
            }
#else
            toonCoord = calcuToon(normalize(outNormal));
#endif
        vec3 toonColor = texture(toonTexture, toonCoord).rgb;
        color.rgb *= toonColor * toonTextureBlend.rgb;
#endif
    
    outPixelColor = color;
}

